/*
 * Space.java
 * Copyright (C) 1998 ymnk, JCraft Inc.
 *
 * This file is part of JSpringies, a mass and spring simulation system for Java
 *
 * JSpringies is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * JSpringies is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with XSpringies; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package com.jcraft.jspringies;

class Space{
  private final static double DT_MIN = 0.0001;
  private final static double DT_MAX = 0.5;
  private final static double CONSTR_KS = 1.0;
  private final static double CONSTR_KD = 1.0;
  private final static double DEF_TSTEP = 0.025;
  private final static int BOUNCE_FUDGE = 1;
  private final static int MAXCON = 1024;
  private final static double STICK_MAG = 1.0;

  public double width, height;

  double cur_mass=1.0, cur_rest=1.0;
  double cur_ks=1.0, cur_kd=1.0;
  boolean fix_mass=false, show_spring=true;
  int center_id=-1;
  int spthick = 0;

  double cur_visc=0.0, cur_stick=0.0;
  double cur_dt=DEF_TSTEP, cur_prec=1.0;
  boolean adaptive_step=false, grid_snap=false;
  double cur_gsnap=20.0;

  Wall wall=null;
  Gravity gravity=null;
  PointerAttraction pattraction=null;
  CenterOfMass cmass=null;
  
  Spring[] springs=null;
  Mass[] masses=null;
  public Mass center_mass=null;

  int num_since = 0;
  double time_elapsed = 0.0;

  double center_x = 0.0;
  double center_y = 0.0;
  double center_rad = 0.0;

  public Space(double w, double h){
    this.width=w; this.height=h;	

//    center_x = width/2.0; center_y = height/2.0;
//    center_rad = 1.0;

    gravity=new Gravity(this, 10.0, 0.0);
    cmass=new CenterOfMass(this, 5.0, 2.0);
    pattraction=new PointerAttraction(this, 10.0, 0.0);
    wall=new Wall(this, 10000.0, 1.0);
  }

  public void clear() { 
    masses=null; 
    springs=null;
  }

  public void addMass(int id, double x, double y, 
              double vx, double vy, double mass, double elastic){
    Mass m=new Mass(id, x, y, vx, vy, mass, elastic);
    if(masses==null){
      masses=new Mass[1];
      masses[0]=m;
    }
    else{
      Mass[] foo=new Mass[masses.length+1];
      System.arraycopy(masses, 0, foo, 1, masses.length);
      foo[0]=m;
      masses=foo;
    }
  }

  public void addSpring(int id, Mass mass1, Mass mass2,
              double ks, double kd, double restlen) {
    Spring spring=new Spring(id, mass1, mass2, ks, kd, restlen);
    if(springs==null){
      springs=new Spring[1];
      springs[0]=spring;
    }
    else{
      Spring[] foo=new Spring[springs.length+1];
      System.arraycopy(springs, 0, foo, 1, springs.length);
      foo[0]=spring;
      springs=foo;
    }
  }

  void accumulate_accel() {
    double gval, gmisc;
    double gx = 0.0, gy = 0.0, ogx = 0.0, ogy = 0.0;

    center_x = width/2.0;
    center_y = height/2.0;
    center_rad = 1.0;

    int i;

    if(masses==null) return;

    if(center_mass!=null){
      if(center_mass.Alivep()){
        center_x = center_mass.x;
        center_y = center_mass.x;
      }
    }

    // Do gravity
    if (gravity.getEnable()){
      gx = gravity.gx;
      gy = gravity.gy;
    }

    if (cmass.getEnable()){
      cmass.force();
      ogx=cmass.ogx;
      ogy=cmass.ogy;
    }
    
    // Apply Gravity, CM and air drag to all masses
    for(i=0; i<masses.length; i++){
      Mass m = masses[i];
      if (m.AliveNotFixedp()) {
        /* Do viscous drag */
        if (m.Id() != center_id) {
          m.ax = gx + ogx - cur_visc * m.vx;
          m.ay = gy + ogy - cur_visc * m.vy;
         } else {
          m.ax = gx - cur_visc * m.vx;
          m.ay = gy - cur_visc * m.vy;
         }
      }
    }

    // Do point attraction force
    if (pattraction.getEnable()){
      pattraction.force();
    }

    // Wall attract/repel force
    if (wall.getEnable()){
      wall.force();
    }

    // Spring compression/damping effects on masses
    if(springs!=null){
      for(i=0; i<springs.length; i++){
	Spring s=springs[i];
	Mass m1, m2;
	if (s.Alivep()) {
	  double dx, dy, force, forcex, forcey, mag, damp, mass1, mass2;
	  m1=s.mass1;
	  m2=s.mass2;    
	  dx = m1.x - m2.x;
	  dy = m1.y - m2.y;
	    
	  if (dx!=0.0 || dy!=0.0) {
	    mag = Math.sqrt(dx * dx + dy * dy);
	    force = s.ks * (s.restlen - mag);

	    //System.out.println("    mag,force="+mag+", "+force);

	    if (s.kd!=0.0) {
	      damp = ((m1.vx - m2.vx) * dx + (m1.vy - m2.vy) * dy) / mag;
	      force -= (s.kd * damp);
	    }
		
	    force /= mag;
	    forcex = force * dx;  forcey = force * dy;
	    mass1 = m1.mass;  mass2 = m2.mass;
		
	    m1.ax += (forcex / mass1);
	    m1.ay += (forcey / mass1);
	    m2.ax -= (forcex / mass2);
	    m2.ay -= (forcey / mass2);
	  }
        }
      }
    }
  }

  void runge_kutta(double h, boolean testloc) {

    int i;

    if(masses==null) return;

    accumulate_accel();

    // k1 step
    for(i=0; i<masses.length; i++){
      Mass m = masses[i];
      if (m.AliveNotFixedp()) {
        m.save2Cur();

        m.k1x = m.vx * h; m.k1y = m.vy * h;
        m.k1vx = m.ax * h; m.k1vy = m.ay * h;
	    
        m.x = m.cur_x + m.k1x / 2.0; m.y = m.cur_y + m.k1y / 2.0;
        m.vx = m.cur_vx + m.k1vx / 2.0;  m.vy = m.cur_vy + m.k1vy / 2.0;
      }
    }
    accumulate_accel();

    // k2 step
    for(i=0; i<masses.length; i++){
      Mass m = masses[i];
      if (m.AliveNotFixedp()) {
        m.k2x = m.vx * (h/2.0);
        m.k2y = m.vy * (h/2.0);
        m.k2vx = m.ax * (h/2.0);
        m.k2vy = m.ay * (h/2.0);
	    
        m.x = m.cur_x + m.k2x / 2.0;
        m.y = m.cur_y + m.k2y / 2.0;
        m.vx = m.cur_vx + m.k2vx / 2.0;
        m.vy = m.cur_vy + m.k2vy / 2.0;
      }
    }
    accumulate_accel();

    // k3 step
    for(i=0; i<masses.length; i++){
      Mass m = masses[i];
      if (m.AliveNotFixedp()) {
        m.k3x = m.vx * (h/2.0);
        m.k3y = m.vy * (h/2.0);
        m.k3vx = m.ax * (h/2.0);
        m.k3vy = m.ay * (h/2.0);
	    
        m.x = m.cur_x + m.k3x;
        m.y = m.cur_y + m.k3y;
        m.vx = m.cur_vx + m.k3vx;
        m.vy = m.cur_vy + m.k3vy;
      }
    }

    accumulate_accel();

    // k4 step
    for(i=0; i<masses.length; i++){
      Mass m = masses[i];
      if (m.AliveNotFixedp()) {
        m.k4x = m.vx * h;
        m.k4y = m.vy * h;
        m.k4vx = m.ax * h;
        m.k4vy = m.ay * h;
      }
    }

    // Find next position
    for(i=0; i<masses.length; i++){
      Mass m = masses[i];
      if (m.AliveNotFixedp()) {
        if (testloc) {
          m.test_x = m.cur_x + (m.k1x/2.0 + m.k2x + m.k3x + m.k4x/2.0)/3.0;
          m.test_y = m.cur_y + (m.k1y/2.0 + m.k2y + m.k3y + m.k4y/2.0)/3.0;
          m.test_vx = m.cur_vx + (m.k1vx/2.0 + m.k2vx + m.k3vx + m.k4vx/2.0)/3.0;
          m.test_vy = m.cur_vy + (m.k1vy/2.0 + m.k2vy + m.k3vy + m.k4vy/2.0)/3.0;
        }
        else {
          m.x = m.cur_x + (m.k1x/2.0 + m.k2x + m.k3x + m.k4x/2.0)/3.0;
          m.y = m.cur_y + (m.k1y/2.0 + m.k2y + m.k3y + m.k4y/2.0)/3.0;
          m.vx = m.cur_vx + (m.k1vx/2.0 + m.k2vx + m.k3vx + m.k4vx/2.0)/3.0;
          m.vy = m.cur_vy + (m.k1vy/2.0 + m.k2vy + m.k3vy + m.k4vy/2.0)/3.0;
        }
      }
    }
  }

/*
  void adaptive_runge_kutta() {
    int i;
    double err, maxerr;

    while(true){
      if (cur_dt > DT_MAX) cur_dt = DT_MAX;
      if (cur_dt < DT_MIN) cur_dt = DT_MIN;

      runge_kutta(cur_dt/2.0, false);
      runge_kutta(cur_dt/2.0, true);

      for(i=0; i<masses.length; i++){
	Mass m = masses[i];
	if (m.AliveNotFixedp()) {
	  m.x = m.old_x; m.y = m.old_y;
	  m.vx = m.old_vx; m.vy = m.old_vy;
	}
      }
      runge_kutta(cur_dt, false);

      // Find error
      maxerr = 0.00001;
      for(i=0; i<masses.length; i++){
	Mass m = masses[i];
	if (m.AliveNotFixedp()) {
	  err = Math.abs(m.x - m.test_x) + Math.abs(m.y - m.test_y) +
	    Math.abs(m.vx - m.test_vx) + Math.abs(m.vy - m.test_vy);
	  if (err > maxerr) maxerr = err;
	}
      }

      // Fudgy scale factor -- user controlled
      maxerr /= cur_prec;
      
      if (maxerr < 1.0) {
	cur_dt *= 0.9 * Math.exp(-Math.log(maxerr)/8.0);
      } 
      else {
	if (cur_dt > DT_MIN) {
	  for(i=0; i<masses.length; i++){
	    Mass m = masses[i];
	    if (m.AliveNotFixedp()) {
	      m.x = m.old_x;
	      m.y = m.old_y;
	      m.vx = m.old_vx;
	      m.vy = m.old_vy;
	    }
	  }
	  cur_dt *= 0.9 * Math.exp(-Math.log(maxerr)/4.0);
	  adaptive_runge_kutta();
	  continue;
	}
      }
      break;
    }
  }
*/

  public boolean animate_obj() {
    int i;
    double stick_mag;
    if(masses==null) return false;

    for(i=0; i<masses.length; i++){
      Mass m = masses[i];
      if (m.AliveNotFixedp()) { 
         m.save2Old();
      }
    }

/*
    if (adaptive_step) {
      boolean any_spring = false;
      for(i=0; i<springs.length; i++){
	Spring s=springs[i];
        if (s.Alivep()){
 	  any_spring = true;
          break;
        }
      }
      if (any_spring) { adaptive_runge_kutta(); }
      else { runge_kutta(cur_dt = DEF_TSTEP, false); }
    } 
    else {
      runge_kutta(cur_dt, false); 
    }
*/

    runge_kutta(cur_dt, false); 

    stick_mag = STICK_MAG * cur_dt * cur_stick;
    double stick=STICK_MAG * cur_stick;

    for(i=0; i<masses.length; i++){
      Mass m = masses[i];
      if (m.AliveNotFixedp()) {
/*
        if (m.ax - m.ax != 0.0 || m.ay - m.ay != 0.0 || 
            m.x - m.x != 0.0 || m.y - m.y != 0.0) {
          //delete_mass(i);
          continue;
        }
*/
	// Check if stuck to a wall
	if (m.old_vx == 0.0 && m.old_vy == 0.0) {
	  if ((wall.left && Math.abs(m.old_x - m.radius) < 0.5) || 
	      (wall.right && Math.abs(m.old_x - width + m.radius) < 0.5)) {
	    if (Math.abs(m.vx) < stick_mag / m.mass) {
	      m.vx = m.vy = 0.0;
	      m.x = m.old_x;
	      m.y = m.old_y;
	      continue;
	    }
	  } 
	  else if ((wall.bottom && Math.abs(m.old_y - m.radius) < 0.5) ||
		   (wall.top && Math.abs(m.old_y - height + m.radius) < 0.5)) {
	    if (Math.abs(m.vy) < stick_mag / m.mass) {
	      m.vx = m.vy = 0.0;
	      m.x = m.old_x;
	      m.y = m.old_y;
	      continue;
	    }		    
	  }
	}

	if (wall.left && m.x < m.radius && m.old_x >= m.radius) {
	  m.x = m.radius;
	  if (m.vx < 0) {
	    m.vx = -m.vx * m.elastic;
	    m.vy *= m.elastic;
	    if (m.vx > 0) {
	      m.vx -= stick / m.mass;
	      if (m.vx < 0) {
		m.vx = m.vy = 0;
	      }
	    }
	  }
	} 
	else if (wall.right && m.x > width - m.radius && 
		 m.old_x <= width - m.radius) {
          m.x = width - m.radius;
          if (m.vx > 0) {
            m.vx = -m.vx * m.elastic;
	    m.vy *= m.elastic;
	    if (m.vx < 0) {
              m.vx += stick / m.mass;
              if (m.vx > 0) {
                m.vx = m.vy = 0;
              }
	    }
	  }
        }
	
	/* Stick to top or bottom wall */
	if (wall.top && m.y < m.radius && m.old_y >= m.radius) {
	  m.y = m.radius;
	  if (m.vy < 0) {
	    m.vy = -m.vy * m.elastic;
	    m.vx *= m.elastic;
	    if (m.vy > 0) {
	      m.vy -= stick / m.mass;
	      if (m.vy < 0) {m.vx = m.vy = 0; }
	    }
	  }
	} 
	else if (wall.bottom && m.y > (height - m.radius) && 
		 m.old_y <= (height - m.radius)) {
          m.y = height - m.radius;
	  if (m.vy > 0) {
	    m.vy = -m.vy * m.elastic;
	    m.vx *= m.elastic;
	    if (m.vy < 0) {
  	      m.vy += stick / m.mass;
              if (m.vy > 0) { m.vx = m.vy = 0; }
	    }
	  }
        }
      }
    }
/*
    time_elapsed += cur_dt;

    if (time_elapsed > 0.05) {
      time_elapsed -= 0.05;
      num_since = 0;
      return true;
    }
    num_since++;
    if (num_since > 8) {
      num_since = 0;
      return true;
    }
*/
    return false;
  }
}
